unit Main;

{$I definitions.inc}

////////////////////////////////////////////////////////////////////////////////
// Casse-brique
// GPL license
// auteur : Matthieu GIROUX
// www.liberlog.fr
////////////////////////////////////////////////////////////////////////////////

interface

uses
  zglSpriteEngine, zengl_sprite_image,
  {$IFNDEF STATIC}
  zglHeader
  {$ELSE}
  zgl_main,
  zgl_screen,
  zgl_window,
  zgl_timers,
  zgl_keyboard,
  zgl_render_2d,
  zgl_fx,
  zgl_textures,
  zgl_textures_png,
  zgl_textures_jpg,
  zgl_sprite_2d,
  zgl_primitives_2d,
  zgl_font,
  zgl_text,
  zgl_math_2d,
  zgl_utils,
  zgl_Mouse,
  zgl_sound_wav,
  zgl_sound
  {$ENDIF}
  ;


{ Zen GL Methodes }
procedure Quit;
procedure Draw;
procedure Init;
procedure TimerGame;

{ Procdures de jeu appele par le timer }
procedure SceneFlipping;
procedure SceneMain;

{ Procdure de nouveau jeu }
procedure StartSceneMain;
{ Procdure de fin de jeu }
procedure EndSceneMain;

{ Constantes }

const
{$IFNDEF FPC}
       DirectorySeparator = '\';
{$ENDIF}
       {Constantes globales }
       ScoreVie    =1000;
       AjouteVie   = 16;
       EcranX=640;
       EcranY=400;

       {Constantes lies aux niveaux }
       Fin = 50;
       {Constantes lies  l'cran }
       BitCompte = 16;
       LimiteXImages = 100 ; {94;}
       LimiteYImages = 50 ;{47;   }

       {Constantes lies aux balles }
       Lente       =20; //Vitesse lente
       VitesseMax = 60;
       ZBalles     =3;
       MaxiBalles   = 60;
       DecalX      = 2;
       DecalY      = 2;
       {Constantes lies aux briques }
       ZBRiques   = 1;
       ZItems     = 0;
       BriquesX   = 8;
       BriquesY   = 30;
       NeCasse    = 1;
       Meurt      = 2;

       TempsChange = 30000;
       TempsAccelere = 20000;
       TempsInversion=5000;
       TempsImmobile=3000;

       {Constantes lies au joueur }
       ZJoueurs    = 2;
       // Joueur 1
       JoueurBasX    = EcranX/2-40;
       JoueurBasY    = EcranY-30;

       BougePaletteX= 30;

       {Constantes lies aux sprites }
       PCasse     = 100;
       Vitesse = 20;
       Ralenti = 18;

       dir_Images = 'Images' + DirectorySeparator;
       dir_Menus = 'Menus' + DirectorySeparator;
       dir_Niveaux = 'Niveaux' + DirectorySeparator;
       dir_Data = 'DATA' + DirectorySeparator ;
       dir_Items = 'Items' ;


       NbBriquesX = 20; {12;}
       NbBriquesY = 20; {14;}
       FinCouche = 5;

       BriqueMinX = 32;
       BriqueMinY = 20 ;
       BBase      = 1;
       TempsEnvoi = 600;
       BSpeedUp   = 8;
       BSpeedDown = 9;
       BIndestructible = 2;

       FinBriques = 28;

       ImageBalle = 'balle' ;
       ImageBrique= 'brick' ;
       ImagePalette= 'palette' ;
       ImageExtension = '.png' ;
       SonExtension = '.wav';

       MaxBriques = 28 ;

       MaxCouche  = 3;

       MOVECOUNT_WITHOUT_TIME = 1 / 100 ;



type  TIndice = 0..3;
      TCases          =ARRAY[0..MaxCouche] of TIndice;
  TInputType  = (itInactive, itMouse,itKeyboard1,itKeyboard2,itJoystick1,itJoystick2);
  TBalleModif = ( bmBalleDessus, bmAccelere, bmRalenti, bmPlomb, bmColle, bmNormal,bmFolleHautBas,bmFolle,bmFolleGaucheDroite,bmPlusGrosse,bmMoinsGrosse);
  TPlayerModif = ( pmErased, pmEnlever, pmAfficher, pmEnvoiBalle, pmBouge );
  TModeBrique = ( mbAucun, mbMeurt );
  UneTouche = array [ 0..0 ] of Byte;


var
  FondImage   : zglPTexture = nil;
  ScoreAjouteVie,
  Score                     : Cardinal ;
  NbLevel,
  Compteur,
  NbBalles,
  NbPalettes,
  Vies,
  PointsVie              : Integer ;
  Stop        : Boolean = False;
  GamePause   : Boolean = False;
  LostGame  : Boolean = False;
  NextLevel : Boolean = False;
  MessageEcran  : string = '';
  SpriteEngine: TSpriteEngine = nil;
  ControleJoueur : TInputType;
  NbJoueurs : Integer;
  fntMain : zglPFont ;
  NbBriques : Integer ;
  Sound_Perdu         ,
  Sound_Casse         ,
  Sound_SpeedUp       : zglPSound;
  zgl_time      : LongWord;
  Presentes     ,
  BrickItems    ,
  Bricks        ,
  Players       ,
  Balls         : TCollisionArray;
  ToucheHaut : UneTouche ;
  ToucheBas  : UneTouche ;
  ToucheGauche : UneTouche ;
  ToucheDroite  : UneTouche ;
  ToucheBoutonA  : UneTouche ;
  ToucheBoutonB  : UneTouche ;


implementation

uses TypInfo, SysUtils;

procedure loadsounds ;
begin
 SubDirSounds := 'Sons' + DirectorySeparator;
 Sound_SpeedUp       := LoadSoundFromRes ('SpeedUp' + SonExtension,1);
 Sound_Casse         := LoadSoundFromRes ('Casse' + SonExtension,1);
 Sound_Perdu         := LoadSoundFromRes ('Perdu' + SonExtension,1);
end;

procedure Init ;
  var
    i : Integer;
begin

  // EN: Create zglCSEngine2D object.
  SpriteEngine := TSpriteEngine.Create();

  snd_Init();
  loadsounds ;

  fntMain := font_LoadFromFile( dirRes + 'font.zfi' );
  StartSceneMain;
end;

type
  TBalle = class;

  { TCollisionSprite }

  TCollisionSprite = class(TAnimatedSprite)
  private
  public
    procedure CollisionBalle ( const Balle : TBalle; const AllowPlomb : Boolean ); virtual;
  end;

   // Sprite du ou des joueur
  { TPlayerSprite }

  TPlayerSprite = class(TAnimatedSprite)
   private
    Deplacement1,Deplacement2,Deplacement3 :Extended;
    BalleModif : TBalleModif;
    ControlX,ControlY,
    Mode,NoPalette : Integer;   // Mode (mort vivant touch)
    Controle : TInputType ;
    PlayerModifs : set of TPlayerModif;
    Intervalle, // Intervalle de rebondissment des balles
    CompteEnvoi,
    CompteurInversion,
    CompteurTire,
    CompteImmobile,
    PaletteEnCours,
    Inversion,
    VitesseBalles : integer;
    NoJoueur : Integer ;
    procedure Erase;
    procedure EnvoiBalle;
    procedure UneCollisionX ( const Sprite : TSprite ; var Done: Boolean);
    procedure ChangeVitesses( const Vitesse : integer);
    function InitPos ( const TheCoord : Double ; const NewCoord : Double ):Double;
    procedure Deplace(const InputCoord: Double);
   public
    constructor Create( const AParent: TSpriteEngine; const Apparaitre : Boolean; const Controle : TInputType); overload; virtual;
    destructor Destroy; override;
    // Methodes abstraites
    procedure BallesModif; overload; virtual; abstract;
    procedure SetPalette; virtual;
    procedure CreatePlayer; virtual; abstract;

    // Methodes implmentes
    procedure DoMove(MoveCount: Double); override;
    procedure DetruitJoueur; virtual;
    procedure Cree; virtual;
    procedure BallesModif ( const Modif : TBalleModif ); overload; dynamic;
    procedure UnMouvement; dynamic;
    procedure UnMouvementX; dynamic;
    procedure BallePlus; dynamic;
    procedure SetBalleColle ( const UneBalle : TBalle ); virtual;
    procedure SetBalleCollePosition ( const UneBalle : TBalle ); virtual;
    procedure Controleur(Control : TInputType); virtual;
    procedure AppuieGaucheDroite     ( const ToucheAEnfoncerGauche   , ToucheAEnfoncerDroite   : Byte ); virtual;
    procedure AppuieButton ( const ToucheAEnfoncer : Byte ); virtual;
    procedure ErasePlayer (); virtual;
  end;


  { TBalle }

  TBalle = class(TAnimatedSprite)
  private
    DecalePalette,
    Vitesse,
    Accelere,
    NoBalle,
    BougeX,BougeY,
    Change   :integer;
    Mode :integer;
    Activite : set of TBalleModif;// Toute l'activit de la balle
    Grosseur : Integer ;
    function PerdJoueur   ( const Perdu : Boolean ):boolean;
    procedure Limites ( var Bouge : Integer );
    procedure ChangeVitesse;
    procedure BalleModif;
    procedure SetImageGrosseur;
    procedure SetImagePosition;

     // Gestion des collisions
  public
    procedure DoMove(MoveCount: Double); override;
    procedure ChangeUneDirection ( const Decalage : Double ); virtual;
    procedure BeforeCollision ( const Sprite: TCollideSprite ); override;
    procedure DoCollision(const Sprite: TCollideSprite; var Done: Boolean); override;
    procedure ColleBalle;
    procedure CollisionBalle ( const Balle : TBalle ); virtual;
    procedure Envoi; virtual;
    constructor Create ( const AParent: TSpriteEngine; Const X, Y : Double ); overload;
    destructor Destroy; override;
    procedure Show ; virtual;
  // Dtruit la balle
    procedure Destruction;
  end;

  ABalle = ^TBalle;

  { TPlayerBas }

  TPlayerBas = class(TPlayerSprite)
  private
  public
    procedure SetBalleColle ( const UneBalle : TBalle ); override;
    procedure UnMouvement; override;
    procedure DoMove(MoveCount: Double); override;
    procedure SetBalleCollePosition ( const UneBalle : TBalle ); override;
    // Gestion des collisions
    procedure DoCollision ( const Sprite: TCollideSprite; var Done: Boolean); override;
    procedure SetPalette; override;
    procedure CreatePlayer; override;
    destructor Destroy; override;
  end;

  { TBrique }

  TBrique = class(TCollisionSprite)
  private
   Cases : TCases;
//  BDGH   : integer;
  NoBrique : Array [ 0 .. MaxCouche ] of TIndice;
  Mode : TModeBrique;
  procedure Touche;
  procedure Destruction;
  procedure DoMove(MoveCount: Double); override;
  public
    procedure CollisionBalle ( const Balle : TBalle ); overload;
    procedure InitBrique ( const DesCases : TCases );
    constructor Create ( const AParent: TSpriteEngine; const I,J : Integer; const DesCases : TCases ); overload;
    destructor Destroy ; override;
  end;

{ TCollisionSprite }

procedure TCollisionSprite.CollisionBalle(const Balle: TBalle; const AllowPlomb : Boolean );
var MilieuX,MilieuY, VecteurAjouteY, VecteurAjouteX                  :integer;
    ModifX, ModifY                  : boolean;
begin
  ModifX:=False;
  ModifY:=False;
  MilieuX:=trunc(Balle.X+Balle.H/2);
  MilieuY:=trunc(Balle.Y+Balle.H/2);
//   begin
  if not AllowPlomb
  or not (bmPlomb in Balle.Activite) then
    begin
      if ((X>=MilieuX) and (Balle.BougeX>0))
      or ((X+W<=MilieuX) and (Balle.BougeX<0)) then
       //La balle vient de la gauche ou de la droite
        Begin
         ModifX:=True;
         VecteurAjouteY := trunc ( X - MilieuX );
        End ;
      if ((Y>=MilieuY) and (Balle.BougeY>0))
      or ((Y+H<=MilieuY) and (Balle.BougeY<0))
       //La balle vient du haut et du bas
       then
        Begin
          ModifY:=True;
          VecteurAjouteX := trunc (  Y - MilieuY );
        End;
    End;

 if ModifX then Balle.BougeX:=-Balle.BougeX;
 if ModifY then Balle.BougeY:=-Balle.BougeY;
 if ModifX and ModifY then
   Begin
    inc( Balle.BougeY, VecteurAjouteY );
    inc( Balle.BougeX, VecteurAjouteX );
   End;
end;

type  UneBrique      = RECORD
                       Brique : TBrique;
                       Existe : Boolean;
                       End;
      TBriques        = ARRAY[0..NbBriquesX-1,0..NbBriquesY-1] of UneBrique;

var
    joueur: TPlayerBas;
    Briques : Tbriques;




procedure TplayerSprite.Erase ();
  Begin
    if not Visible then
      Begin
        CanDoCollisions := False ;
        Visible := False;
        dec(NbPalettes);
      End;
  End ;


procedure TplayerSprite.AppuieButton ( const ToucheAEnfoncer : Byte );
  Begin
    if ([ pmEnlever, pmAfficher ] * PlayerModifs <> [] )
      Then
        Exit ;
    case Controle of
      itMouse :
        begin
          if (mouse_Click ( 0 )) and self.Visible then
             begin
                PlayerModifs := PlayerModifs + [pmBouge];
                EnvoiBalle;
              end;
         End;
      itKeyboard1,itKeyboard2 :
         begin
           if key_Press ( ToucheAEnfoncer) and self.Visible then
             begin
                PlayerModifs := PlayerModifs + [pmBouge];
                EnvoiBalle;
              end;
          End;
      End;
  end;

constructor TPlayerSprite.Create(const AParent: TSpriteEngine; const Apparaitre : Boolean; const Controle : TInputType );
begin
  inherited Create(AParent);
  Cree;
  SetPalette;
  Controleur(Controle);
  RegisterCollideArray ( Players );
end;

destructor TPlayerSprite.Destroy;
begin
  UnRegisterCollideArray ( Players );
  inherited Destroy;
end;

procedure TPlayerSprite.ErasePlayer ();
begin
  PlayerModifs := PlayerModifs + [pmEnlever];
  Controleur ( itInactive );
end;


procedure TPlayerSprite.DoMove(MoveCount: Double);
begin
  AppuieButton ( ToucheBoutonA [ NoJoueur ] );
  Erase ;
  if CompteImmobile=0 then
   begin
     UnMouvement;
   end
    else dec(CompteImmobile);
  inherited Domove ( MoveCount );
  If ( Deplacement1 > 0 ) then
    Begin
     PlayerModifs := PlayerModifs + [pmBouge];
    End;
end;


procedure TPlayerSprite.Deplace(const InputCoord: Double);
begin
 Deplacement3:=Deplacement2;
 Deplacement2:=Deplacement1;
 Deplacement1:=InputCoord*Inversion;
end;

procedure TPlayerSprite.DetruitJoueur;
begin
  if (NbJoueurs>0)
  and (NBBalles=0)
  and ( [pmBouge,pmEnlever] * PlayerModifs = [] )
   Then
    Begin
      Setlength ( Balls, 0 );
      Controleur ( itInactive );
      PlayerModifs := PlayerModifs + [pmEnlever,pmErased];
    End;
end;
function TPlayerSprite.InitPos ( const TheCoord : Double ; const NewCoord : Double ): Double;
begin
  Deplacement3:=0;
  Deplacement2:=0;
  Deplacement1:=NewCoord-trunc(TheCoord);
  Result:=NewCoord;
End;

procedure TPlayerSprite.UnMouvementX;
begin
 Deplace(ControlX);
 if X+W+Deplacement1+Deplacement2 / 4+Deplacement3 / 8>EcranX then
  begin
    X := InitPos ( X, EcranX-W );
  end
  else
 if X+Deplacement1+Deplacement2 / 4+Deplacement3 / 8<0 then
  begin
    X := InitPos ( X, 0 );
  End
   else
   X:=X+Deplacement1+Deplacement2 / 4+Deplacement3 / 8;
End;
procedure TplayerSprite.AppuieGaucheDroite     (  const ToucheAEnfoncerGauche   , ToucheAEnfoncerDroite : Byte  );
  begin
    if (( [pmEnlever,pmAfficher] * PlayerModifs ) = [] ) then
      case Controle of
        itMouse : begin
                 ControlX:=trunc(mouse_DX-X);
                 End;
        itKeyboard1,itKeyboard2 :
                 begin
                 if key_Press ( ToucheAEnfoncerDroite) then
                 ControlX:=round ( 10+Deplacement2 / 2+ Deplacement3 / 3 )
                  else
                  if key_Press ( ToucheAEnfoncerGauche) then
                   ControlX:=round ( -10+Deplacement2 / 2 + Deplacement3 / 3 )
                    else ControlX:=0;
                 End;
        end;
  end;

procedure TPlayerSprite.SetPalette;
begin
 case NoPalette of
  3 : SetImage ( ImagePalette );
  End;
 If (( pmAfficher in PlayerModifs ) and Visible and ( PaletteEnCours = NoPalette )) Then
   Begin
     Controle:= ControleJoueur ;
     PlayerModifs:=PlayerModifs-[pmAfficher];
   End;
end;

procedure TPlayerSprite.Cree;
begin
  PlayerModifs := [pmAfficher];
  inc(NbPalettes);
  CompteurTire := 0 ;
  CompteurInversion:=0;
  CompteEnvoi:=0;
  Inversion:=1;
  NoPalette:=3;
  PaletteEnCours:=1;
  CompteImmobile:=0;
  VitesseBalles :=Lente;
  Mode:=0;
  Deplacement1:=0;
  Deplacement2:=0;
  Deplacement3:=0;
  Mode:=0;
//  AnimSpeed := Vitesse/1000;
  Visible:=True;
  CanDoCollisions:=True;
  CreatePlayer ;
End;


procedure TPlayerSprite.Controleur(Control : TInputType);
begin
  if ( Control <> Controle ) Then
    Begin
      Controle:=Control;
      if Controle=itInactive then
        begin
         Intervalle:=0;
         PlayerModifs:=PlayerModifs+[pmEnlever];
         Dec(NbPalettes);
         Visible:=False;
    //       CanDoMoving:=False;
         CanDoCollisions:=False;
    //       EnvoiBalle;
        End
       Else
        CanDoCollisions:=True;
    End;
end;

procedure TPlayerSprite.BallePlus;   // balle supplmentaire
var UneBalle     : TBalle ;
begin
  if NbBalles > Maxiballes Then
    Exit;
  UneBalle := TBalle.Create(SpriteEngine, Self.X + Self.W / 2, Self.Y + Self.H / 2 );
End;

procedure TPlayerSprite.ChangeVitesses(const Vitesse : integer);
begin
  Randomize;
  VitesseBalles:=Vitesse+Random(VitesseMax-Vitesse);
End;


procedure TPlayerSprite.BallesModif ( const Modif : TBalleModif );
begin
  BalleModif:=Modif;
End;


procedure TPlayerSprite.UnMouvement;
begin
  inc (CompteurTire);
  If CompteEnvoi=TempsEnvoi then EnvoiBalle;
  if CompteurInversion>0 then dec(CompteurInversion)
   else Inversion:=1;
  inc(CompteEnvoi);
  inc(Compteur);
  if (NoPalette<>PaletteEnCours) and (Compteur mod 5=0) then
   begin
     if PaletteEncours<NoPalette then inc(PaletteEnCours);
     if PaletteEncours>NoPalette then dec(PaletteEnCours);
     SetPalette;
    //  end;
   End;
end;

procedure TPlayerSprite.UneCollisionX(const Sprite: TSprite; var Done: Boolean);
var UneBalle : TBalle ;
begin
  if (Sprite is TBalle) then // and (Y+Milieu>=JoueurBasY) then
   begin
     UneBalle := Tballe(Sprite);
     if not ( bmColle in UneBalle.Activite ) Then
       Begin
         UneBalle.BougeY:=-UneBalle.BougeY ;
         CompteEnvoi:=0;
         UneBalle.Change:=0;
         UneBalle.DecalePalette:= trunc( Uneballe.X + UneBalle.W/2 - Self.X - Self.W / 2 );
         if ( Abs ( UneBalle.DecalePalette ) > Self.W / 3 ) Then
           if ( UneBalle.BougeX <> 0 ) Then
                  UneBalle.BougeX := UneBalle.BougeX div UneBalle.BougeX * UneBalle.DecalePalette
             Else UneBalle.BougeX := UneBalle.DecalePalette ;
         UneBalle.DecalePalette:= trunc ( Uneballe.X - Self.X );
         UneBalle.Limites(UneBalle.BougeX);
         UneBalle.DoMove(MOVECOUNT_WITHOUT_TIME );
        End;
     UneBalle.ColleBalle;
   end;
end;


procedure TPlayerSprite.SetBalleColle(const UneBalle: TBalle);
begin
  UneBalle.ColleBalle;
end;

procedure TPlayerSprite.SetBalleCollePosition(const UneBalle: TBalle);
begin
end;


procedure TPlayerSprite.EnvoiBalle;

begin
  CompteEnvoi:=0;
  PlayerModifs := PlayerModifs + [pmEnvoiBalle];
End;



procedure TPlayerBas.SetPalette;
begin
  inherited;

end;

procedure TPlayerBas.UnMouvement;
begin
  UnMouvementX;
  inherited;
end;

procedure TPlayerBas.DoMove (MoveCount: Double);
begin
  AppuieGaucheDroite ( ToucheGauche [ NoJoueur ], ToucheDroite [ NoJoueur ] );
  inherited;
end;

procedure TPlayerBas.DoCollision ( const Sprite: TCollideSprite; var Done: Boolean);
begin
  UneCollisionX(Sprite,Done);
  inherited DoCollision( Sprite, Done);
end;

procedure TPlayerBas.CreatePlayer;
var I       :Integer;
begin
  X := JoueurBasX;
  Y:= JoueurBasY;
  NoJoueur := 0 ;
End ;

destructor TPlayerBas.Destroy;
begin
  joueur:=nil;
  inherited Destroy;
end;


procedure TPlayerBas.SetBalleColle(const UneBalle: TBalle);
begin
  inherited SetBalleColle(UneBalle);
  UneBalle.DecalePalette:= trunc ( UneBalle.X - Self.x );
end;

procedure TPlayerBas.SetBalleCollePosition(const UneBalle: TBalle);
begin
  UneBalle.X:=self.X+UneBalle.DecalePalette;
  UneBalle.Y:=self.Y-UneBalle.W;
end;



procedure TBrique.Destruction;
var i : Integer;
begin
  PlaySoundRandom ( Sound_Casse );
  dec(NbBriques);
  for i := 0 to MaxCouche - 1 Do
    Begin
      NoBrique[i]:=NoBrique [ I + 1 ];
      if ( NoBrique [ I + 1 ] = 0 ) Then
        Break;
    End ;
  NoBrique[high ( NoBrique )]:= 0 ;
  if NoBrique[0]=0 then
   begin
     inc(Score,PCasse*NbPalettes);
     Dead;
   end
    else
     Begin
       SetImage(  ImageBrique +IntToStr(NoBrique[0]));
       Mode:=mbAucun;
       Frame:=0;
       AnimActive:=False;
       CanDoCollisions:=True;
     End;
   if NbBriques=0 then NextLevel:=True;
end;

procedure TBrique.Touche;
begin
  Mode:=mbMeurt;
  CanDoCollisions:=False;
end;

procedure TBrique.DoMove(MoveCount: Double);
begin
  if Mode = mbMeurt Then
    Destruction;
  inherited;
end;

procedure TBrique.CollisionBalle(const Balle: TBalle);
begin
  CollisionBalle ( Balle, True );
  Touche ;
  if ( Mode = mbMeurt )
   then
    Balle.Accelere:=0;
end;

procedure TBrique.InitBrique( const DesCases : TCases );
var k : Integer ;
begin
  For K:=0 to FinCouche do
   Begin
     Cases[k]:=DesCases[k];
     if (Cases[K]>0) and (Cases[K]<=FinBriques) then
      begin
        NoBrique[K]:=0;
        if ( Cases[k]<=FinBriques ) then
          NoBrique[K]:=Cases[K];
        inc(NbBriques);
      End
       Else
         Break;

   end;
end;

//Cre une brique
constructor TBrique.Create ( const AParent: TSpriteEngine; const I,J : Integer; const DesCases : TCases );
var Indice : TIndice ;
begin
  inherited Create(AParent);
  Indice := DesCases [ 0 ];
  Mode:=mbAucun;
  SetImage(  ImageBrique +IntToStr(Indice));
  NoBrique[0]:=Indice;
  Visible:=True;
  CanDoCollisions:=True;
//  AnimCount := Texture.FramesX;
  AnimActive := False;
  InitBrique(DesCases);
  RegisterCollideArray ( Bricks );
  X:=BriquesX+I*W;
  Y:=BriquesY+J*H;
//  AnimSpeed := Vitesse/1000;
end;

destructor TBrique.Destroy;
begin
  inherited Destroy;
  UnRegisterCollideArray ( Bricks );
end;

procedure Tballe.SetImageGrosseur;
begin
  SetImage (ImageBalle);
  SetImagePosition;
End ;

procedure Tballe.SetImagePosition;
begin
  if (bmPlomb in Activite) Then
    Frame := 1
   Else
    Frame := 0 ;
End ;

procedure Tballe.BalleModif;
begin

case joueur.BalleModif of
 bmNormal :
   begin
     Activite := [];
     SetImagePosition;
   end;
 bmPlomb :
   Begin
     Activite := Activite + [bmPlomb];
     SetImagePosition;
   end;
 bmFolle :
   Begin
     Activite := Activite + [bmFolle];
   end;
 bmColle :
   Begin
     Activite := Activite + [bmColle];
   end;
 bmFolleHautBas :
   Begin
     Activite := Activite + [bmFolleHautBas];
   end;
 bmFolleGaucheDroite :
   Begin
     Activite := Activite + [bmFolleGaucheDroite];
   end;
 bmPlusGrosse :
   Begin
    inc ( Grosseur );
    If Grosseur > 3 Then
      Grosseur := 3
     Else
      SetImageGrosseur;
   end;
 bmMoinsGrosse :
   Begin
    dec ( Grosseur );
    If Grosseur < 1 Then
      Grosseur := 1
     Else
      SetImageGrosseur;
   end;
 End;

End;

procedure Tballe.ChangeVitesse;
Begin
  if not ( bmColle in Activite ) Then
    Begin
      if BougeX > 0 Then BougeX:=trunc(Vitesse*(BougeX/BougeX))
                    Else BougeX:=Vitesse;
      if BougeY > 0 Then BougeY:=trunc(Vitesse*(BougeY/BougeY))
                    Else BougeY:=Vitesse;
      if BougeX > VitesseMax Then BougeX := VitesseMax;
      if BougeY > VitesseMax Then BougeY := VitesseMax;
    End;
End;

procedure Tballe.Limites ( var Bouge : Integer );
Begin
  if (Abs(Bouge)>=VitesseMax) then Bouge:=VitesseMax*Bouge div Bouge;
End;

procedure TBalle.Destruction;
Begin
  dec(NBBalles);
  //MainForm.Donnee:=intToStr(NBBalles);
  if NBBalles<=0 then LostGame:=True;
  Frame:=0;
  Activite:=[];
  Visible:=False;
  CanDoMoving:=False;
  CanDoCollisions:=False;
End;

procedure TBalle.CollisionBalle(const Balle: TBalle);
var Dirige : Integer ;
begin
 if  not ( bmColle in Activite )
 and not ( bmBalleDessus in Activite ) then
  begin
    Activite := Activite + [bmBalleDessus];
    Randomize;
    Dirige:=Random(2);
    if Dirige<>0 then BougeX:=-Balle.BougeX*Dirige;
    if Dirige<>0 then BougeY:=-Balle.BougeY*Dirige;
  // if BougeX<0 then BougeX:=BougeX-Dirige;
  // if BougeY<0 then BougeY:=BougeY-Dirige;
    Limites(BougeX);
    Limites(BougeY);
  end;

end;


// gestion des collisions
procedure TBalle.BeforeCollision ( const Sprite: TCollideSprite);
begin
   if (Sprite is TBalle   ) Then TestCollision := TestCollisionCircleCircle
                            else TestCollision := TestCollisionCircleRectangle;
end;

// gestion des collisions
procedure TBalle.DoCollision ( const Sprite: TCollideSprite; var Done : Boolean);
begin
  if (Sprite is TBrique) then // si la balle rencontre une brique
    (Sprite as TBrique ).CollisionBalle ( Self )
  else if (Sprite is TBalle) then
     CollisionBalle ( Sprite as TBalle )
   else
     Activite := Activite - [bmBalleDessus];
end;

procedure TBalle.ColleBalle;
begin
  if ( joueur.BalleModif = bmColle ) Then
    Begin
      CanDoCollisions:=False;
      Activite := Activite + [bmColle];
      BougeX := 0;
      BougeY := 0;
    End;
end;

Function TBalle.PerdJoueur ( const Perdu : Boolean ):boolean;
Begin
  Result:=False;
  if Perdu
  and Visible then
    begin
     Destruction;
     Result:=True;
     joueur.DetruitJoueur;
    End;
End;

procedure TBalle.DoMove(MoveCount: Double);
begin
  if ( X < 0 ) or ( X + W > EcranX ) Then
    BougeX := -BougeX;
  if ( Y < 0 ) Then
    BougeY := -BougeY;
  if  ( BougeX = 0 )
  and ( BougeY = 0 )
  and not ( bmColle in Activite )
    Then
     Begin
       if X > EcranX / 2 Then
          BougeX := - Lente
         Else
          BougeX := Lente ;
       if Y > EcranY / 2 Then
          BougeY := - Lente
         Else
          BougeY := Lente ;
     End ;

  inc(Accelere);

  if Accelere>TempsAccelere then
   begin
     PlaySoundRandom ( Sound_Casse, False );
     if ( bmRalenti in Activite )
       Then Activite := Activite + [bmAccelere] - [bmRalenti]
       Else Activite := Activite - [bmAccelere] + [bmRalenti];
     Accelere:=0;
     if ( bmRalenti in Activite ) and (Vitesse>1) then dec(Vitesse)
      Else if (Vitesse<VitesseMax) then inc(Vitesse);
     ChangeVitesse;
   End;
  if (assigned ( Joueur )) Then
    Begin
      BalleModif;
      Vitesse := Joueur.VitesseBalles;
      if bmColle in Activite Then
        Begin
          if ( pmEnvoiBalle in Joueur.PlayerModifs )
             Then Envoi
             Else Joueur.SetBalleCollePosition ( Self );
        End ;
      End ;
  PerdJoueur (Y>EcranY);
  if Change>=TempsChange then // au bout d'un moment la balle change de direction
   if ( bmColle in Activite ) Then
      Envoi
     Else
       begin
         Change:=0;
         Randomize;
         Activite:=Activite-[bmColle];
         if (Abs(BougeX)>=VitesseMax)
          then
           BougeX:=Lente * ( Random ( 1 ) + 2 );
         Randomize;
         if (Abs(BougeY)>=VitesseMax)
          then
           BougeY:=Lente * ( Random ( 1 ) + 2 );
       End;
  X:=X+BougeX*MoveCount;    // Dplacement de la itMouse
  Y:=Y+BougeY*MoveCount;
  Inc(Change);  // Compteur qui sert  contrler la direction de la balle
   // Collisions ensuite
  inherited DoMove(MoveCount);
//  Collision;
end;

// Change la direction de la balle
// En entre la soustractions des Y aux X
procedure TBalle.ChangeUneDirection(const Decalage: Double);
begin
  If ( Decalage < 0 ) Then
    BougeY:=-BougeY
   Else
    BougeX:=-BougeX ;
end;

procedure TBalle.Envoi;
begin
 CanDoCollisions := True;
 BougeY:=-lente;
 BougeX:=-lente;
 Activite:=Activite-[bmColle];
end;

//Cre une brique
constructor TBalle.Create ( const AParent: TSpriteEngine; Const X, Y : Double );
begin
  inherited Create(AParent);
  Mode:=0;
  Activite := [bmColle] ;
  Grosseur := 1 ;
  Vitesse:=Lente;
  DecalePalette := 0 ;
  BougeX := 0 ;
  BougeY := 0 ;
  Accelere:=0;
  SetImageGrosseur;
  SetImagePosition;
  AnimActive := False;
  Visible:=False;
  Self.NoBalle:=NbBalles;
  Show;
  CanDoCollisions:=False;
  Self.X := X ;
  Self.Y := Y ;
  Joueur.SetBalleColle ( Self );
  RegisterCollideArray ( Balls );
end;

destructor TBalle.Destroy;
begin
  inherited Destroy;
  UnRegisterCollideArray ( Balls );
end;

procedure TBalle.Show;
begin
  If not Visible Then
    Begin
      Visible := True ;
      inc ( NbBalles );
    End;

end;


procedure Quit;
begin
  // RU:     .
  // EN: Free allocated memory for sprites.
  SpriteEngine.Clear ;
  SpriteEngine.Free;
  SpriteEngine := nil;
end;

procedure Draw;
begin
  batch2d_Begin();
  // RU:        .
  // EN: Render all sprites contained in current sprite engine.
    //My code here
  SceneFlipping;

  if time > 255 Then
    SpriteEngine.Draw();


  if time > 255 Then
    begin
      pr2d_Rect( 0, 0, 256, 64, $000000, 200, PR2D_FILL );
      text_Draw( fntMain, 0, 0, 'FPS: ' + u_IntToStr( zgl_Get( RENDER_FPS ) ) );
      text_Draw( fntMain, 0, 20, 'Sprites: ' + u_IntToStr( SpriteEngine.Count ) );
      text_Draw( fntMain, 0, 40, 'Up/Down - Add/Delete Miku :)' );
    end;
  batch2d_End();
end;
// Rafraichissement d'cran
procedure TimerGame;
begin
  INC( zgl_time, 2 );

  if key_Press( K_ESCAPE )
    Then
     begin
       zgl_Exit();
       Stop := True;
     end;
  key_ClearState();

  if Stop Then
    Exit;

  // RU:        .
  // EN: Process all sprites contained in current sprite engine.
  SpriteEngine.Proc();
  // RU:      .
  // EN: Delete all sprites if space was pressed.
  //  Fenetre:=FMenu;
  Draw;
End ;

// Rafraichissement d'cran
procedure SceneFlipping;

//var Rect : TAdRect ;
var i :integer;
begin
  if NextLevel then
    begin
      Score:=Score;
      Stop := True ;
//      DXTimer.Enabled:=False;
      EndSceneMain;
      StartSceneMain;
      Stop := False ;
//      DXTimer.Enabled:=True;
    end;
  if LostGame then
    begin
      dec(Vies);
      LostGame:=False;
      if Vies>0 then
       begin
         NbBalles:=0;
         if not ( [pmErased,pmEnlever,pmAfficher] * joueur.PlayerModifs <> [] ) then
          begin
            joueur.Erase;
            joueur.Cree;
            joueur.BallePlus;
          End;
       End;
      if Vies<=0 then
         begin
           PlaySoundRandom ( Sound_Perdu,False);
           Score:=0;
           EndSceneMain;
           StartSceneMain;
         end;
    End;
  SceneMain;
  mouse_ClearState;
  SpriteEngine.Draw;

end;

// Dmarrage du niveau      ---------------------------------
procedure StartSceneMain;

var

//  NoSprite
  i, j , k           :Integer;
  Cases : TCases;

begin
 EndSceneMain;
 NbLevel:=0;
 Score:=0;
 PointsVie:=0;
 ScoreAjouteVie:=ScoreVie;
 NbLevel:=0;
 Vies:=10;

  Stop := True ;
    ControleJoueur := itMouse;
    {  Main scene beginning  }
    inc(NbLevel);
    if NbLevel>Fin then NbLevel:=1;

    SubImages:= dir_Images ;
    NbBalles:=0;
    NbBriques :=0;
    NbPalettes:=0;
    NextLevel:=False;
    Randomize;
    for I:=0 to NbBriquesX-1 do for J:=0 to NbBriquesY - 1 do
     begin
       Cases[0]:=random(3);
       // couches supplmentaires de briques si il existe une brique avant
       for K:=1 to MaxCouche do
         Begin
           if Cases[K-1] = 0 Then
             Break ;
           inc (NbBriques);
           Cases[K]:=random(3);
         end;
       if Cases[MaxCouche]>0 Then
         inc (NbBriques);
       // Cration de la premire brique
       if (Cases[0]>0) and (Cases[0]<=FinBriques) then
        begin
          Briques[I,J].Brique:=TBrique.Create(SpriteEngine,I,J,Cases );
        End
         else
         Briques[I,J].Brique:=nil;
     end;
    joueur:= TPlayerBas.Create(SpriteEngine, True, ControleJoueur ) ;
    joueur.BallePlus;
    if NbBriques=0 then NextLevel:=True;

  Stop:= False;
end;



// Fin de l'animation destruction des sprites
procedure EndSceneMain;
begin
  Stop:=True;
    {  Main scene end  }
    //destruction des sprites
  SpriteEngine.Clear ;
  Finalize ( Presentes );
  Finalize ( Bricks   );
  Finalize ( Players  );
  Finalize ( Balls    );

end;

// Scne d'animation
procedure SceneMain;
var I : Integer ;
begin

  If  ((( ControleJoueur = itMouse    ) and ( mouse_Click ( 0 ) ))
  or  ((( ControleJoueur = itKeyboard1  ) or ( ControleJoueur = itKeyboard2  ))))
  and ( pmErased in joueur.PlayerModifs     )
    Then
      joueur.PlayerModifs := joueur.PlayerModifs + [pmAfficher] ;//AnimOuvreBas ;

  SpriteEngine.Collide(BrickItems,Players);
  SpriteEngine.Collide(Balls     ,Players);
  SpriteEngine.Collide(Balls     ,Bricks );
  SpriteEngine.Collide(Balls     ,Presentes );
  SpriteEngine.Collide(Players   ,Presentes );
  SpriteEngine.Move( MOVECOUNT_WITHOUT_TIME);
  SpriteEngine.Draw;

  if Compteur<100 then
    text_Draw( fntMain, 10, 30, MessageEcran)
   else MessageEcran:='';
  inc(Compteur);
end;

finalization
end.

